# (C) Crown Copyright, Met Office. All rights reserved.
#
# This file is part of cube_helper and is released under the
# BSD 3-Clause license.
# See LICENSE in the root of the repository for full licensing details.
__version__ = '2.1.0'

from cube_helper.cube_help import (load,
                                   add_categorical,
                                   aggregate_categorical,
                                   extract_categorical,
                                   concatenate,
                                   extract)
from cube_helper.cube_loader import (load_from_dir,
                                     load_from_filelist,
                                     sort_by_earliest_date,
                                     file_sort_by_earliest_date)
from cube_helper.cube_equaliser import (examine_dim_bounds,
                                        equalise_time_units,
                                        equalise_attributes,
                                        equalise_dim_coords,
                                        equalise_aux_coords,
                                        equalise_data_type,
                                        equalise_all,
                                        remove_attributes,
                                        compare_cubes)
from cube_helper.fix_known import fix_known_issues
from cube_helper.logger import (muffle_logger,
                                reset_logger)



# (C) Crown Copyright, Met Office. All rights reserved.
#
# This file is part of cube_helper and is released under the
# BSD 3-Clause license.
# See LICENSE in the root of the repository for full licensing details.

import os
import iris
import glob
from iris.exceptions import MergeError, ConstraintMismatchError
from six import string_types
from datetime import datetime


def _check_pdt_year(cell, partial_datetime):
    if partial_datetime.year:
        return partial_datetime.year
    else:
        return cell.point.year


def _check_pdt_month(cell, partial_datetime):
    if partial_datetime.month:
        return partial_datetime.month
    else:
        return cell.point.month


def _check_pdt_day(cell, partial_datetime):
    if partial_datetime.day:
        return partial_datetime.day
    else:
        return cell.point.day


def _check_pdt_hour(cell, partial_datetime):
    if partial_datetime.hour:
        return partial_datetime.hour
    else:
        return cell.point.hour


def _check_pdt_minute(cell, partial_datetime):
    if partial_datetime.minute:
        return partial_datetime.minute
    else:
        return cell.point.minute


def _check_pdt_second(cell, partial_datetime):
    if partial_datetime.second:
        return partial_datetime.second
    else:
        return cell.point.second


def _check_pdt_microsecond(cell, partial_datetime):
    if partial_datetime.microsecond:
        return partial_datetime.microsecond
    else:
        return cell.point.microsecond


def _fix_partial_datetime(constraint):
    if isinstance(constraint._coord_values['time'], iris.time.PartialDateTime):
        part_datetime = constraint._coord_values['time']
        new_constraint = iris.Constraint(
            time=lambda cell:
            cell.point.year == _check_pdt_year(cell, part_datetime) and
            cell.point.month == _check_pdt_month(cell, part_datetime) and
            cell.point.day == _check_pdt_day(cell, part_datetime) and
            cell.point.hour == _check_pdt_hour(cell, part_datetime) and
            cell.point.minute == _check_pdt_minute(cell, part_datetime) and
            cell.point.second == _check_pdt_second(cell, part_datetime) and
            cell.point.microsecond ==
            _check_pdt_microsecond(cell, part_datetime))
        return new_constraint
    else:
        return constraint


def _constraint_compatible(constraint, cube):
    try:
        cube.extract(constraint)
        return True
    except (TypeError, ConstraintMismatchError):
        return False


def _parse_directory(directory):
    """
    Parses the string representing the directory, makes sure a '/'
    backspace is present at the start and end of the directory string
    so glob can work properly.

    Args:
         directory: the directory string to parse.

    Returns:
        a string representing the directory, having been parsed if
        needed.
    """
    if not directory.endswith('/'):
        directory = directory + '/'

    if not directory.startswith('/'):
        if os.path.isdir(directory):
            return directory

        else:
            directory = '/' + directory
            return directory
    else:
        return directory


def _sort_by_date(time_coord):
    """
    Private sorting function used by _file
    _sort_by_earliest_date() and sort_by_earl
    iest_date().

    Args:
        time_coord: Cube time coordinate for each cube
        to be sorted by.

    Returns:
        time_origin: The time origin to sort cubes
        by, as a specific start date e.g 1850.
    """
    time_origin = time_coord.units.num2date(0)
    if not isinstance(time_origin, datetime):
        if time_origin.datetime_compatible:
            time_origin = time_origin._to_real_datetime()
        else:
            time_origin = datetime(time_origin.year,
                                   time_origin.month,
                                   time_origin.day)
    return time_origin


def file_sort_by_earliest_date(cube_filename):
    """
    Sorts file names by date from earliest to latest.

    Args:
        cube_filename: list of files in string format to sort,
        to be used with CubeList sort method when cube_load is called.

    Returns:
        datetime object of selected Cubes start time.
    """
    raw_cubes = iris.load_raw(cube_filename)
    if isinstance(raw_cubes, iris.cube.CubeList):
        for cube in raw_cubes:
            if isinstance(cube.standard_name, string_types):
                for time_coord in cube.coords():
                    if time_coord.units.is_time_reference():
                        time_origin = _sort_by_date(time_coord)
                        return time_origin
    else:
        for time_coord in iris.load_cube(cube_filename).coords():
            if time_coord.units.is_time_reference():
                time_origin = _sort_by_date(time_coord)
                return time_origin


def sort_by_earliest_date(cube):
    """
    Sorts Cubes by date from earliest to latest.

    Args:
        cube: CubeList or list to sort, to be used with CubeList
        sort method when cube_load is called.

    Returns:
        datetime object of selected Cubes start time.
    """

    for time_coord in cube.coords():
        if time_coord.units.is_time_reference():
            time_origin = _sort_by_date(time_coord)
            return time_origin


def load_from_dir(directory, filetype, constraint=None):
    """
    Loads a set of cubes from a given directory, single cubes are loaded
    and returned as a CubeList.

    Args:
        directory: a chosen directory
        to operate on. directory MUST start and end with forward
        slashes.

        filetype (optional): a string specifying the expected type
        Of files found in the dataset.

        constraints (optional): a string specifying any constraints
        You wish to load the dataset with.

    Returns:
        iris.cube.CubeList(loaded_cubes), a CubeList of the loaded
        Cubes.
    """
    if constraint is None:
        loaded_cubes = []
        cube_files = []
        directory = _parse_directory(directory)
        cube_paths = glob.glob(directory + '*' + filetype)
        for path in cube_paths:
            try:
                loaded_cubes.append(iris.load_cube(path))
                cube_files.append(path)
            except (MergeError, ConstraintMismatchError):
                for cube in iris.load_raw(path):
                    if isinstance(cube.standard_name, str):
                        loaded_cubes.append(cube)
                        cube_files.append(path)
        loaded_cubes.sort(key=sort_by_earliest_date)
        cube_files.sort(key=file_sort_by_earliest_date)
        return loaded_cubes, cube_files
    else:
        loaded_cubes = []
        cube_files = []
        directory = _parse_directory(directory)
        cube_paths = glob.glob(directory + '*' + filetype)
        if not _constraint_compatible(constraint,
                                      iris.load_cube(cube_paths[0])):
            constraint = _fix_partial_datetime(constraint)
        for path in cube_paths:
            try:
                loaded_cubes.append(iris.load_cube(path, constraint))
                cube_files.append(path)
            except (MergeError, ConstraintMismatchError):
                for cube in iris.load_raw(path, constraint):
                    if isinstance(cube.standard_name, str):
                        loaded_cubes.append(cube)
                        cube_files.append(path)
        loaded_cubes.sort(key=sort_by_earliest_date)
        cube_files.sort(key=file_sort_by_earliest_date)
        return loaded_cubes, cube_files


def load_from_filelist(paths, filetype, constraint=None):
    """
    Loads the specified files. Individual files are
    returned in a
    CubeList.

    Args:
        paths: a chosen list of filenames to operate on.

        filetype (optional): a string specifying the expected type
        Of files found in the dataset

        constraints (optional): a string, iterable of strings or an
        iris.Constraint specifying any constraints you wish to load
        the dataset with.

    Returns:
        iris.cube.CubeList(loaded_cubes), a CubeList of the loaded
        Cubes.
    """
    loaded_cubes = []
    cube_files = []
    for filename in paths:
        if not filename.endswith(filetype):
            paths.remove(filename)

    for filename in paths:
        if constraint is None:
            try:
                loaded_cubes.append(iris.load_cube(filename))
                cube_files.append(filename)
            except (MergeError, ConstraintMismatchError):
                for cube in iris.load_raw(filename):
                    if isinstance(cube.standard_name, str):
                        loaded_cubes.append(cube)
                        cube_files.append(filename)

        else:
            if not _constraint_compatible(constraint,
                                          iris.load_cube(paths[0])):
                constraint = _fix_partial_datetime(constraint)
            try:
                loaded_cubes.append(iris.load_cube(filename, constraint))
                cube_files.append(filename)
            except (MergeError, ConstraintMismatchError):
                for cube in iris.load_raw(filename, constraint):
                    if isinstance(cube.standard_name, str):
                        loaded_cubes.append(iris.load_raw(filename,
                                                          constraint))
                        cube_files.append(filename)
    loaded_cubes.sort(key=sort_by_earliest_date)
    cube_files.sort(key=file_sort_by_earliest_date)
    return loaded_cubes, cube_files



# (C) Crown Copyright, Met Office. All rights reserved.
#
# This file is part of cube_helper and is released under the
# BSD 3-Clause license.
# See LICENSE in the root of the repository for full licensing details.

from __future__ import (absolute_import, division, print_function)
import sys
import numpy as np
from iris.util import unify_time_units
from collections import namedtuple
from itertools import combinations
from cube_helper.logger import log_module, log_inconsistent, log_coord_remove


def equalise_attributes(cubes, comp_only=False):
    """
    Equalises Cubes for concatenation and merging, cycles through the
    iterable of cubes (either a list of loaded cubes or aan iris CubeList)
    and removes any that are not common across all cubes, metadata,
    and variables.

    Args:
        cubes: Cubes to be equalised of attributes.

        comp_only: A boolean value, if set to True it will examine
        the cubes attributes and print inconsistencies
        but not equalise them.

    Returns:
        Equalised cube_dataset to the CubeHelp class

    """
    uncommon_keys = set()
    attr_dict_list = []
    attr_dict = {}
    for cube in cubes:
        for key, value in cube.attributes.items():
            if isinstance(value, np.ndarray):
                value = str(value)
                attr_dict.update({key: value})
            else:
                attr_dict.update({key: value})
        attr_dict_list.append(attr_dict.items())
        attr_dict = {}
    combs = list(combinations(attr_dict_list, 2))
    for element in combs:
        for key in set(element[0]) ^ set(element[1]):
            uncommon_keys.add(key[0])

    if not comp_only:
        for key in uncommon_keys:
            for cube in cubes:
                try:
                    del cube.attributes[key]
                except KeyError:
                    pass
        log_coord_remove(list(uncommon_keys), 'attributes')
    else:
        log_inconsistent(list(uncommon_keys), 'attibutes')
    return cubes


def equalise_time_units(cubes, comp_only=False):
    """
    Equalises time units by cycling through each cube in the given CubeList
    or list of loaded cubes.

    Args:
        cubes: Cubes to equalised of time coords.

        comp_only: A boolean value, if set to True it will examine
        the cubes time_coordinates and print inconsistencies but not
        equalise them.

    Returns:
        cubes with time coordinates unified.
    """
    logger = log_module()
    comp_messages = set()
    change_messages = set()
    calendar = cubes[0].coord('time').units.calendar
    origin = cubes[0].coord('time').units.origin
    for cube in cubes:
        for time_coord in cube.coords():
            if time_coord.units.is_time_reference():
                if comp_only:
                    if time_coord.units.calendar != calendar:
                        comp_messages.add("\tcalendar format inconsistent\n")
                    if time_coord.units.origin != origin:
                        comp_messages.add("\ttime start date inconsistent\n")
                else:
                    if origin != time_coord.units.origin:
                        change_messages.add("New time origin set to "
                                            "{}\n".format(origin))
                        unify_time_units(cubes)
    if comp_messages:
        for message in comp_messages:
            logger.info(message)
    if change_messages:
        for message in change_messages:
            logger.info(message)

    return cubes


def equalise_data_type(cubes, data_type='float32'):
    """
    Casts datatypes in iris numpy array to be of the same datatype.

    Args:
        cubes: Cubes to have their datatypes equalised.

        data_type: String specifying datatype, default is float32.


    Returns:
        cubes: Cubes with their data types identical.
    """
    logger = log_module()
    if data_type == 'float32':
        for cube in cubes:
            cube.data = np.float32(cube.data)
    elif data_type == 'float64':
        for cube in cubes:
            cube.data = np.float64(cube.data)
    elif data_type == 'int32':
        for cube in cubes:
            cube.data = np.int32(cube.data)
    elif data_type == 'int64':
        for cube in cubes:
            cube.data = np.int64(cube.data)
    else:
        logger.error("invalid data type")


def equalise_dim_coords(cubes, comp_only=False):
    """
    Equalises dimensional coordinates of Cubes, specifically long_name,
    standard_name, and var_name.

    Args:
        cubes: CubeList or list of Cubes to equalise.

        comp_only: A boolean value, if set to True it will examine
        the cubes dimension coordinates and print inconsistencies
        but not equalise them.

    Returns:
        Cubes equalised across dimension coordinates.
    """
    logger = log_module()
    inconsistency_sn = set()
    inconsistency_ln = set()
    inconsistency_vn = set()
    inconsistency_attr = set()
    coord_dict = {}
    for cube in cubes:
        for coord in cube.dim_coords:
            coord_dict.update(
                {coord.name(): {'long_name': coord.long_name,
                                'standard_name': coord.standard_name,
                                'var_name': coord.var_name,
                                'attributes': coord.attributes}})

    for coord in coord_dict:
        for cube in cubes:
            if comp_only:
                if cube.coord(coord).standard_name != \
                        coord_dict[coord]['standard_name']:
                    inconsistency_sn.add(coord)
                if cube.coord(coord).long_name != \
                        coord_dict[coord]['long_name']:
                    inconsistency_ln.add(coord)
                if cube.coord(coord).var_name != \
                        coord_dict[coord]['var_name']:
                    inconsistency_vn.add(coord)
                if cube.coord(coord).attributes != \
                        coord_dict[coord]['attributes']:
                    inconsistency_attr.add(coord)
            else:
                try:
                    cube.coord(coord).standard_name = \
                        coord_dict[coord]['standard_name']
                    cube.coord(coord).long_name = \
                        coord_dict[coord]['long_name']
                    cube.coord(coord).var_name = \
                        coord_dict[coord]['var_name']
                    cube.coord(coord).attributes = \
                        coord_dict[coord]['attributes']
                except ValueError:
                    pass
    if any([inconsistency_sn,
            inconsistency_ln,
            inconsistency_vn,
            inconsistency_attr]):
        logger.info("\ncube dim coordinates differ: \n")
    if comp_only:
        log_inconsistent(list(inconsistency_sn), 'coords standard_name')
        log_inconsistent(list(inconsistency_ln), 'coords long_name')
        log_inconsistent(list(inconsistency_vn), 'coords var_name')
        log_inconsistent(list(inconsistency_attr), 'coords attributes')
    return cubes


def equalise_aux_coords(cubes, comp_only=False):
    """
    Equalises auxillary coordinates of cubes.

    Args:
        cubes: CubeList or list of Cubes to equalise.

        comp_only: A boolean value, if set to True it will examine
        the Cube's aux coordinates and print inconsistencies
        but not equalise them.

    Returns:
        Cubes equalised across auxillary coordinates.
    """
    logger = log_module()
    inconsistencies = set({})
    change_messages = set({})
    cube_combs = list(combinations(cubes, 2))
    for combs in cube_combs:
        cube_a_dict = {c.name(): c for c in combs[0].aux_coords}
        cube_b_dict = {c.name(): c for c in combs[1].aux_coords}
        cube_a_coords = {c for c in cube_a_dict}
        cube_b_coords = {c for c in cube_b_dict}
        for coord in list(cube_a_coords):
            if coord not in cube_b_coords:
                if comp_only:
                    inconsistencies.add(coord)
                elif coord == 'height':
                    change_messages.add("Adding {} coords to cube\n".
                                        format(coord))
                    combs[1].add_aux_coord(cube_a_dict[coord])
        for coord in list(cube_b_coords):
            if coord not in cube_a_coords:
                if comp_only:
                    inconsistencies.add(coord)
                elif coord == 'height':
                    change_messages.add("Adding {} coords to cube\n".
                                        format(coord))
                    combs[0].add_aux_coord(cube_b_dict[coord])
    if inconsistencies:
        inconsistencies = list(inconsistencies)
        log_inconsistent(inconsistencies, 'coords')

    if change_messages:
        for message in change_messages:
            logger.info(message)

    return cubes


def remove_attributes(cubes):
    """
    Sets all cube attributes to an empty string.
    be unaffected.

    Args:
        cubes: Cubes to have attributes stripped.

    Returns:
        Cubes with attributes replaced with empty string ''.
    """
    for cube in cubes:
        for attr in cube.attributes:
            cube.attributes[attr] = ''


def equalise_all(cubes):
    """
    Invokes equalise_aux_coords, equalise_attributes,
    equalise_dim_coords and equalise_time units all at once.
    Used before cube_load concatenates the cubes.

    Args:
        cubes: Cubes to be equalised.

    Returns:
        cubes: Cubes equalised across metadata and coords.


    """
    cubes = equalise_aux_coords(cubes)
    cubes = equalise_attributes(cubes)
    cubes = equalise_dim_coords(cubes)
    cubes = equalise_time_units(cubes)
    return cubes


def compare_cubes(cubes):
    """
    Examines coordinates and attributes across iterable of iris cubes
    And calls equalise functions (with comp_only arg set to true) where
    appropriate.

    Args:
        cubes: An iterable of iris Cubes or CubeList to be compared
        for inconsostencies.

    Returns:
        A printed string detailing the inconsistencies in the cubes.
    """
    logger = log_module()
    uneq_aux_coords = False
    uneq_dim_coords = False
    uneq_attr = False
    uneq_time_coords = False
    uneq_ndim = False
    cube_combs = list(combinations(cubes, 2))
    for comb in cube_combs:
        if (uneq_aux_coords and
                uneq_attr and
                uneq_dim_coords and
                uneq_time_coords and
                uneq_ndim):
            break

        if comb[0].aux_coords != comb[1].aux_coords:
            uneq_aux_coords = True

        if comb[0].dim_coords != comb[1].dim_coords:
            uneq_dim_coords = True

        if comb[0].attributes != comb[1].attributes:
            uneq_attr = True

        if comb[0].ndim != comb[1].ndim:
            uneq_ndim = True
            break

        if comb[0].coord('time').units != comb[1].coord('time').units:
            uneq_time_coords = True
            break

    if uneq_ndim:
        logger.error("Number of dimensions for cubes differ,"
                     " please load cubes of matching ndim")
        raise OSError

    if uneq_aux_coords:
        logger.info("\ncube aux coordinates differ: \n")
        equalise_aux_coords(cubes, comp_only=True)

    if uneq_dim_coords:
        equalise_dim_coords(cubes, comp_only=True)

    if uneq_attr:
        logger.info("cube attributes differ: \n")
        equalise_attributes(cubes, comp_only=True)

    if uneq_time_coords:
        logger.info("cube time coordinates differ: \n")
        equalise_time_units(cubes, comp_only=True)


def _examine_dim_bounds(cubes, cube_files):
    Range = namedtuple('Range', ['start', 'end'])
    msg = ''
    for i, cube_a in enumerate(cubes):
        for j, cube_b in enumerate(cubes):
            if i != j:
                range_a = Range(start=cube_a.coord('time').bounds[0][0],
                                end=cube_a.coord('time').bounds[-1][-1])
                range_b = Range(start=cube_b.coord('time').bounds[0][0],
                                end=cube_b.coord('time').bounds[-1][-1])
                latest_start = max(range_a.start, range_b.start)
                earliest_end = min(range_a.end, range_b.end)
                delta = earliest_end - latest_start
                overlap = max(0, delta)
                if overlap > 0:
                    msg = msg + "\nThe time coordinates overlap at cube {}" \
                                " and cube {}".format(i, j)
                    msg = msg + "\nThese cubes are: \n\t{}\n\t{}"\
                        .format(cube_files[i], cube_files[j])
                    break
    return msg


def examine_dim_bounds(cubes, cube_files):
    """
    Examines the dimensional bounds of time should concatenate fail.
    Cycles through cubes and determines if the times are contiguous.

    Args:
         cubes: Iris cubes to examine the time bounds of

         cube_files: the respective files of cubes, to give users
         info as to what cubes are causing problems with concatentation.

    Returns:
        A printed string detailing any overlap in the time bounds.
    """
    logger = log_module()
    msg = _examine_dim_bounds(cubes, cube_files)
    logger.info(msg)



# -*- coding: utf-8 -*-
# (C) Crown Copyright, Met Office. All rights reserved.
#
# This file is part of cube_helper and is released under the
# BSD 3-Clause license.
# See LICENSE in the root of the repository for full licensing details.
from __future__ import (absolute_import, division, print_function)
import iris
import iris.coord_categorisation
from six import string_types
from cube_helper.logger import log_module, muffle_logger, reset_logger
from cube_helper.cube_loader import (load_from_filelist,
                                     load_from_dir,
                                     _constraint_compatible,
                                     _fix_partial_datetime)
from cube_helper.cube_equaliser import (compare_cubes,
                                        equalise_all,
                                        _examine_dim_bounds)


def load(directory, filetype='.nc', constraints=None):
    """
    A function that loads and concatenates Iris Cubes.

    Args:
        directory: A String specifying the directory or filename of the Cubes
        you wish to concatenate. Accepts either directory location or a list
        or glob object of individual Cubes.

        filetype: Extension of Iris Cubes to Load. set to '.nc' by default.

        constraints: Any constraints to be applied to Cubes on load.

    Returns:
        result: A concatenated Iris Cube.
    """
    logger = log_module()
    if isinstance(directory, string_types):
        loaded_cubes, cube_files = load_from_dir(
            directory, filetype, constraints)
        if not loaded_cubes:
            raise OSError("No cubes loaded")
        else:
            compare_cubes(loaded_cubes)
            result = equalise_all(loaded_cubes)
            result = iris.cube.CubeList(result)
            try:
                result = result.concatenate_cube()
                return result
            except iris.exceptions.ConcatenateError:
                logger.info("\nThere was an error in concatenation\n")
                err_msg = _examine_dim_bounds(result, cube_files)
                logger.error(err_msg)
                raise

    elif isinstance(directory, list):
        loaded_cubes, cube_files = load_from_filelist(
            directory, filetype, constraints)

        if not loaded_cubes:
            raise OSError("No cubes loaded")
        else:
            compare_cubes(loaded_cubes)
            result = equalise_all(loaded_cubes)
            result = iris.cube.CubeList(result)
            try:
                result = result.concatenate_cube()
                return result
            except iris.exceptions.ConcatenateError:
                logger.info("\nThere was an error in concatenation\n")
                err_msg = _examine_dim_bounds(result, cube_files)
                logger.error(err_msg)
                raise


def _season_year(**kwargs):
    iris.coord_categorisation.add_season_year(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='season_year')


def _season_membership(**kwargs):
    iris.coord_categorisation.add_season_membership(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='season_membership',
        season=kwargs.get('season'))


def _season_number(**kwargs):
    iris.coord_categorisation.add_season_number(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='season_number',
        seasons=kwargs.get('seasons'))


def _season(**kwargs):
    iris.coord_categorisation.add_season(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='season',
        seasons=kwargs.get('seasons'))


def _clim_season(**kwargs):
    iris.coord_categorisation.add_season(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='clim_season',
        seasons=kwargs.get('seasons'))


def _year(**kwargs):
    iris.coord_categorisation.add_year(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='year')


def _month_number(**kwargs):
    iris.coord_categorisation.add_month_number(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='month_number')


def _month_fullname(**kwargs):
    iris.coord_categorisation.add_month_fullname(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='month_fullname')


def _month(**kwargs):
    iris.coord_categorisation.add_month(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='month')


def _day_of_month(**kwargs):
    iris.coord_categorisation.add_day_of_month(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='day_of_month')


def _day_of_year(**kwargs):

    iris.coord_categorisation.add_day_of_year(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='day_of_year')


def _weekday_number(**kwargs):
    iris.coord_categorisation.add_weekday_number(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='weekday_number')


def _weekday_fullname(**kwargs):
    iris.coord_categorisation.add_weekday_fullname(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='weekday_fullname')


def _weekday(**kwargs):
    iris.coord_categorisation.add_weekday(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='weekday')


def _hour(**kwargs):
    iris.coord_categorisation.add_hour(
        kwargs.get('cube'),
        kwargs.get('coord'),
        name='hour')


def _annual_seasonal_mean(**kwargs):
    _clim_season(**kwargs)
    _season_year(**kwargs)


def _add_categorical(cube, categorical, coord, season, seasons):
    categorical_dict = {'season_year': _season_year,
                        'season_membership': _season_membership,
                        'season_number': _season_number,
                        'annual_seasonal_mean': _annual_seasonal_mean,
                        'number': _season_number,
                        'clim_season': _clim_season,
                        'season': _season,
                        'year': _year,
                        'month_number': _month_number,
                        'month_fullname': _month_fullname,
                        'month': _month,
                        'day_of_month': _day_of_month,
                        'day_of_year': _day_of_year,
                        'weekday_number': _weekday_number,
                        'weekday_fullname': _weekday_fullname,
                        'weekday': _weekday,
                        'hour': _hour
                        }

    categorical_dict.get(categorical)(cube=cube,
                                      categorical=categorical,
                                      coord=coord,
                                      season=season,
                                      seasons=seasons)


def add_categorical(cubes, categorical, coord='time', season='djf',
                    seasons=('djf', 'mam', 'jja', 'son')):
    """
    Adds a coordinate categorisation(s) to the iterable of Iris Cubes.

    Currently this function provides a wrapper for the following
    standalone and compound categoricals:

    day_of_month:
        Add a day-of-month coordinate, vals 1-31.
    day_of_year:
        Add a day-of-year coordinate, vals 1-365 (366 in leap years).
    month:
        Add a month coordinate, vals ‘Jan’-’Dec’.
    month_fullname:
        Add a month coordinate, vals ‘January’-’December’.
    month_number:
        Add a month coordinate, vals 1-12.
    season:
        Add a season-of-year coordinate, with specified seasons.
    clim_season:
        Add a climatalogical season-of-year coordinate, with
        specified seasons.
    season_membership:
        Add a season membership coordinate for a specified season.
    season_number:
        Add a season-of-year coordinate, values 0..(N-1)
        where N is the number of user specified seasons.
    season_year:
        Add a categorical year-of-season coordinate, with specified seasons.
    weekday:
        Add a weekday coordinate, vals ‘Mon’-’Sun’.
    weekday_fullname:
        Add a weekday coordinate, vals ‘Monday’-’Sunday’.
    weekday_number:
        Add a weekday coordinate, vals 0-6 [0=Monday].
    year:
        Add a calendar-year coordinate.
    annual_seasonal_mean:
        Add a clim_season and a season_year coordinate.

    Args:
        cubes: A cube, a list of Loaded Cubes or a CubeList.

        categorical: A string or list of strings specifying
        the categorisation you wish to add. Additionally a compound
        categorisation can be added. E.g 'annual_seasonal_mean'.

        coords: The coordinate you wish to add a categorisation to. Set
        to 'time' by default.

        season: The season you need for the categorisation (where required).
        set to 'djf' by default.

        seasons: The seasons required for categorisation.

    Returns:
        cubes: An iterable of Cubes, either a list of loaded Cubes
        or an Iris CubeList.
    """
    if isinstance(categorical, list):
        for categorical in categorical:

            if isinstance(cubes, list) or isinstance(cubes,
                                                     iris.cube.CubeList):
                for cube in cubes:
                    _add_categorical(cube, categorical, coord, season, seasons)
            else:
                _add_categorical(cubes, categorical, coord, season, seasons)
        return cubes
    else:
        if isinstance(cubes, list) or isinstance(cubes, iris.cube.CubeList):
            for cube in cubes:
                _add_categorical(cube, categorical, coord, season, seasons)
            return cubes

        else:
            _add_categorical(cubes, categorical, coord, season, seasons)
            return cubes


def aggregate_categorical(cube, categorical,
                          coord='time', season='djf',
                          seasons=('djf', 'mam', 'jja', 'son'),
                          agg_method=iris.analysis.MEAN):
    """
    Adds a coordinate categorisation(s) to the iterable of Iris Cubes, then
    aggregates them by the given categoricals. Categoricals used are the
    same as the ones suppourted by add_categorical().

    Args:
        cubes: A cube, a list of loaded Cubes or a CubeList.

        categorical: A string or list of strings specifying
        the categorisation you wish to add. Additionally a compound
        categorisation can be added. E.g 'annual_seasonal_mean'.

        coord: The coordinate you wish to add a categorisation to. Set
        to 'time' by default.

        season: The season you need for the categorisation (where required).
        set to 'djf' by default.

        seasons: The seasons required for categorisation.

        agg_method: An Iris aggregator object, e.g ``iris.analysis.MEAN``,

    Returns:
        cubes: A cube, a list of loaded Cubes, or an Iris CubeList
        aggregated by a given categorical.
    """
    compound_dict = {'annual_seasonal_mean': ['clim_season',
                                              'season_year']}
    cube = add_categorical(cube, categorical, coord=coord, season=season,
                           seasons=seasons)
    try:
        categorical = compound_dict[categorical]
        cube = cube.aggregated_by(categorical, agg_method)
        return cube
    except KeyError:
        cube = cube.aggregated_by(categorical, agg_method)
        return cube


def extract_categorical(cube,
                        categorical,
                        constraint,
                        coord='time',
                        season='djf',
                        seasons=('djf', 'mam', 'jja', 'son'),
                        agg_method=iris.analysis.MEAN):
    """
    Adds a coordinate categorical, aggregates by said categorical,
    then extracts the given contraint. The categoricals used are the
    then extracts the given contraint. Categoricals used are the
    same as the ones suppourted by add_categorical() and
    aggregate_categorical().

    Args:
        cubes: A cube, a list of Loaded Cubes or a CubeList.

        categorical: A string or list of strings specifying
        the categorisation you wish to add. Additionally a compound
        categorisation can be added. E.g 'annual_seasonal_mean'.

        constraint: An Iris constraint you wish to extract.

        coords: The coordinate you wish to add a categoisation to. Set
        to 'time' by default.

        season: The season you need for the categorisation (where required).
        set to 'djf' by default.

        seasons: The seasons required for categorisation.

    Returns:
        cubes: A cube, a list of loaded Cubes, or an Iris CubeList
        extracted from a given constraint.
    """
    if not isinstance(constraint, iris.Constraint):
        raise NameError("No constraint given")

    else:
        cube = aggregate_categorical(cube,
                                     categorical,
                                     coord,
                                     season,
                                     seasons,
                                     agg_method)
        return cube.extract(constraint)


def concatenate(cubes):
    """
    Concatenates a list of Iris Cubes. Equalises the list of cubes first
    then concatenates.

    Args:
         cubes: An iterable of Iris Cubes to concatenate, either list of
         loaded cubes or a CubeList

    Returns:
        cube: A concatenated Iris Cube.
    """
    cubes = equalise_all(cubes)
    cube_list = iris.cube.CubeList(cubes)
    cube = cube_list.concatenate_cube()
    return cube


def extract(cube, constraint):
    """
    Extracts a constraint on an Iris cube, and will fix common issues
    associated with time constraints and partial datetimes.

    Note: This function will not be able to rectify constraints using
    lambda functions. An extract_bounds function will be added
    to a future release of cube_helper to help with more complex
    time extractions.

    Args:
        cube: A cube or a CubeList.

        constraint: A constraint to extract.

    Returns:
        the cube or CubeList extracted along the constraint.
    """
    if not _constraint_compatible(cube, constraint):
        new_constraint = _fix_partial_datetime(constraint)
        return cube.extract(new_constraint)
    else:
        return cube.extract(constraint)



# (C) Crown Copyright, Met Office. All rights reserved.
#
# This file is part of cube_helper and is released under the
# BSD 3-Clause license.
# See LICENSE in the root of the repository for full licensing details.
import logging
import sys


class CapturableHandler(logging.StreamHandler):

    @property
    def stream(self):
        return sys.stdout

    @stream.setter
    def stream(self, value):
        pass


def _add_handler(logger, level, handler):
    logger.setLevel(level)
    handler = handler
    formatter = logging.Formatter('%(message)s')
    handler.setFormatter(formatter)
    handler.setLevel(level)
    logger.addHandler(handler)
    logger.handler_set = True


def _remove_handler(logger):
    logger.handlers.pop()
    logger.handler_set = False


def log_module():
    logger = logging.getLogger(__name__)
    handler = CapturableHandler()
    if not getattr(logger, 'handler_set', None):
        _add_handler(logger, logging.INFO, handler)
    return logger


def muffle_logger():
    """
    A function that switches the logging level to ``ERROR`` or higher, thereby
    muffling all output other than error messages. As this
    option obscures the user to changes made to cubes on load
    (as well as inconsistencies) it is advised not to use this
    function unless absolutely necessary.
    """
    logger = logging.getLogger(__name__)
    handler = CapturableHandler()
    if getattr(logger, 'handler_set', None):
        _remove_handler(logger)
        _add_handler(logger, logging.ERROR, handler)
    else:
        _add_handler(logger, logging.ERROR, handler)


def reset_logger():
    """
    A function that switches the level to ``INFO`` or higher, allowing
    ``cube_helper``'s logger to detail changes and inconsistencies as
    usual
    """
    logger = logging.getLogger(__name__)
    handler = CapturableHandler()
    if getattr(logger, 'handler_set', None):
        _remove_handler(logger)
        _add_handler(logger, logging.INFO, handler)
    else:
        _add_handler(logger, logging.INFO, handler)


def _to_comma_and_str(component_list, metadata_component):
    if len(component_list) > 1:
        msg = ", ".join(component_list[:-1]) + " and " + component_list[-1]
        return msg + ' ' + metadata_component
    elif len(component_list) == 1:
        msg = component_list[0]
        return msg + ' ' + metadata_component
    else:
        msg = ""
        return str(msg + metadata_component)


def log_inconsistent(component_list, metadata_component):
    logger = log_module()
    if component_list:
        msg = "\t" + \
              _to_comma_and_str(component_list, metadata_component) + \
              " inconsistent\n"
        logger.info(str(msg))


def log_coord_remove(component_list, metadata_component):
    logger = log_module()
    if component_list:
        msg = ""
        for comp in component_list:
            if len(component_list) == 1:
                msg = "{} ".format(comp)
            elif comp != component_list[-1]:
                msg = msg + "{}, ".format(comp)
            else:
                msg = msg + "and {} ".format(comp)
        msg = "Deleting " + \
              _to_comma_and_str(component_list, metadata_component) + \
              " from cubes\n"
        logger.info(str(msg))



# (C) Crown Copyright, Met Office. All rights reserved.
#
# This file is part of cube_helper and is released under the
# BSD 3-Clause license.
# See LICENSE in the root of the repository for full licensing details.
import os
import setuptools


def get_long_description():
    """Use the contents of README.md as the long description"""
    with open("README.md", "r") as fh:
        return fh.read()


def extract_version():
    """
    Retrieve version information from the  __init__.py module.
    """
    version = ''
    directory = os.path.dirname(__file__)
    filename = os.path.join(directory, 'cube_helper', '__init__.py')

    with open(filename) as fd:
        for line in fd:
            line = line.strip()
            if line.startswith('__version__'):
                try:
                    version = line.split('=')[1].strip(' "\'')
                except Exception:
                    pass
                break

    if not version:
        print('WARNING: Unable to parse version information from '
              'file: {}'.format(filename))
        version = '0.0.0'

    return version


setuptools.setup(
    name='cube_helper',
    packages=['cube_helper'],
    version=extract_version(),
    license='BSD 3-Clause License',
    description=('Cube Helper is a package to make equalisation, '
                 'concatenation, and analysis of Iris cubes easier.'),
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author='Met Office',
    author_email='jon.seddon@metoffice.gov.uk',
    url='https://github.com/MetOffice/cube_helper',
    download_url='https://github.com/MetOffice/cube_helper/releases',
    keywords=['climate', 'cmip', 'iris', 'cubes', 'preprocessing'],
    install_requires=[
        'scitools-iris'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering :: Atmospheric Science',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



# (C) Crown Copyright, Met Office. All rights reserved.
#
# This file is part of cube_helper and is released under the
# BSD 3-Clause license.
# See LICENSE in the root of the repository for full licensing details.
from abc import ABCMeta, abstractmethod
import six

import iris

from cube_helper.logger import log_module


@six.add_metaclass(ABCMeta)
class FixKnownIssue:
    """
    An abstract base class to identify and fix known issues with data loaded
    into an `iris.cube.Cube`.

    Args:
        cube: the cube to check, and if necessary fix.
    """
    def __init__(self, cube):
        self.cube = cube
        self.logger = log_module()
        self.message = ''

    @abstractmethod
    def is_fix_needed(self):
        """
        Check if this fix applies to the object's cube.

        Returns:
            True if this fix should be applied to the cube that it was
            instantiated with.
        """
        pass

    @abstractmethod
    def fix_cube(self):
        """
        Apply the fix to the object's cube and inform about the name and
        contents of the fix that has been applied.
        """
        log_message = 'Applying {}. {}'.format(self.__class__.__name__,
                                               self.message)
        self.logger.info(log_message)


@six.add_metaclass(ABCMeta)
class FixKnownIssueIdentifyAttributes(FixKnownIssue):
    """
    An abstract base class to identify and fix known issues with data loaded
    into an `iris.cube.Cube`. A cube that needs this fix is identified by
    the class' `cube_attributes_required` attribute key-value pairs being
    found in `cube.attributes`.

    Args:
        cube: the cube to check, and if necessary fix.
    """
    @abstractmethod
    def __init__(self, cube):
        super(FixKnownIssueIdentifyAttributes, self).__init__(cube)
        self.cube_attributes_required = {}

    def is_fix_needed(self):
        """
        Check if this fix applies to the object's cube by checking that all of
        the key-value pairs in `self.cube_attributes_required` occur in
        `cube.attributes`.

        Returns:
            True if this fix should be applied to the cube that it was
            instantiated with.
        """
        test_result = True
        for key, value in six.iteritems(self.cube_attributes_required):
            if key in self.cube.attributes:
                if self.cube.attributes[key] != value:
                    test_result = False
                    break
            else:
                test_result = False
                break

        return test_result


class FixCmip6CasFgoals(FixKnownIssueIdentifyAttributes):
    """
    Fix the latitude and longitude bounds in

    * mip_era: CMIP6
    * institution_id: CAS
    * source_id: FGOALS-f3-L

    that are not contiguous and use `iris.coords.Coord.guess_bounds()` to
    calculate new bounds for these two coordinates.

    Args:
        cube: the cube to check, and if necessary fix.
    """
    def __init__(self, cube):
        super(FixCmip6CasFgoals, self).__init__(cube)
        self.cube_attributes_required = {
            'mip_era': 'CMIP6',
            'institution_id': 'CAS',
            'source_id': 'FGOALS-f3-L'
        }
        self.message = 'Fixing latitude and longitude bounds.'

    def fix_cube(self):
        """
        Delete the existing faulty bounds on the latitude and longitude
        coordinates and then calculate new bounds.
        """
        super(FixCmip6CasFgoals, self).fix_cube()
        self.cube.coord('latitude').bounds = None
        self.cube.coord('latitude').guess_bounds()
        self.cube.coord('longitude').bounds = None
        self.cube.coord('longitude').guess_bounds()


class FixCmip6FioqlnmFioesm20Latitude(FixKnownIssueIdentifyAttributes):
    """
    Fix the latitude and longitude bounds in

    * mip_era: CMIP6
    * institution_id: FIO-QLNM
    * source_id: FIO-ESM-2-0
    * latitude is an auxiliary coordinate

    which are different to other FIO-ESM-2-0 experiments and have the wrong
    bounds along with a latitude that is not monotonic, preventing latitude
    from being a dimension coordinate. `iris.coords.Coord.guess_bounds()`
    calculates bounds for these two coordinates that are identical to the
    bounds in other experiments,  which also allows latitude to be promoted to
    a dimension coordinate.

    Args:
        cube: the cube to check, and if necessary fix.
    """
    def __init__(self, cube):
        super(FixCmip6FioqlnmFioesm20Latitude, self).__init__(cube)
        self.cube_attributes_required = {
            'mip_era': 'CMIP6',
            'institution_id': 'FIO-QLNM',
            'source_id': 'FIO-ESM-2-0',
        }
        self.message = ('Fixing latitude and longitude bounds and promoting '
                        'latitude to a dimension coordinate.')

    def is_fix_needed(self):
        """
        Check if this fix applies to the object's cube by checking that all of
        the key-value pairs in `self.cube_attributes_required` occur in
        `cube.attributes` and that latitude is an auxiliary coordinate.

        Returns:
            True if this fix should be applied to the cube that it was
            instantiated with.
        """
        attributes_check = super(FixCmip6FioqlnmFioesm20Latitude,
                                 self).is_fix_needed()

        aux_coord_names = [aux_coord.name()
                           for aux_coord in self.cube.aux_coords]
        aux_coord_check = 'latitude' in aux_coord_names

        return attributes_check and aux_coord_check

    def fix_cube(self):
        """
        Delete the existing faulty bounds on the latitude and longitude
        coordinates, calculate new bounds and promote latitude to be a
        dimension coordinate.
        """
        super(FixCmip6FioqlnmFioesm20Latitude, self).fix_cube()
        self.cube.coord('latitude').bounds = None
        self.cube.coord('latitude').guess_bounds()
        self.cube.coord('longitude').bounds = None
        self.cube.coord('longitude').guess_bounds()
        iris.util.promote_aux_coord_to_dim_coord(self.cube, 'latitude')


# List all concrete fixes that should be applied. They should also be
# described in the docstring for `fix_known_issues()`
ALL_FIXES = [FixCmip6CasFgoals, FixCmip6FioqlnmFioesm20Latitude]


def fix_known_issues(cube):
    """
    Identify any issues that are known about in the input cube, fix these
    in-place and warn the user what has been fixed.

    The issues currently fixed are:

    **FixCmip6CasFgoals** *mip_era* `CMIP6` *institution_id* `CAS` *source_id*
    `FGOALS-f3-L`: remove the latitude and longitude bounds, which are not
    contiguous, and use `iris.coords.Coord.guess_bounds()` to calculate new
    bounds for these two coordinates.

    **FixCmip6FioqlnmFioesm20Latitude** *mip_era* `CMIP6` *institution_id*
    `FIO-QLNM` *source_id* `FIO-ESM-2-0` *and* `latitude is an auxiliary
    coordinate`: remove the latitude and longitude bounds, which are invalid,
    and use `iris.coords.Coord.guess_bounds()` to calculate new bounds for
    these two coordinates, which are identical to the bounds in other
    `FIO-ESM-2-0` experiments. latitude is then promoted to a dimension
    coordinate.

    Args:
        cube: the iris.cube.Cube to fix
    """
    for fixer in ALL_FIXES:
        fix = fixer(cube)
        if fix.is_fix_needed():
            fix.fix_cube()
